package ludumdare19.gamestate
{
    import ludumdare19.gameobjects.Player;
    import ludumdare19.gameobjects.TileMapWithUserData;
    import ludumdare19.rooms.RoomData;
    import ludumdare19.rooms.RoomLoader;
    import ludumdare19.rooms.TileSet;
    import ludumdare19.Sfx;
    import stego.constants.Colors;
    import stego.GameInterface;
    import stego.GameObject;
    import stego.gameobject.CanvasGameObject;
    import stego.GameState;
    import stego.ui.DisappearingText;
    import stego.util.collision.CollisionInfo;
    import stego.util.MathUtil;
    import stego.util.scene.GameObjectHolder;
    import stego.util.StringUtil;

    /**
     * ...
     * @author Mike Meyer
     */
    public class Playing extends GameState
    {
        /** layer to put the everything in */
        protected var layer:GameObjectHolder;

        protected var player:Player;
        protected var roomRow:Number;
        protected var roomCol:Number;
        protected var roomLoader:RoomLoader;
        protected var tileMap:TileMapWithUserData;
        protected var noise:CanvasGameObject;

        protected var piecesCollected:Array;

        protected static const collectPieceMessages:Array = [ "Collected one piece out of four.",
                                                              "Two out of four pieces collected.",
                                                              "Just one piece remaining!",
                                                              "All pieces collected!" ];
        static public const NOISE_ALPHA:Number = 0.05;

        /**
         * create scene
         */
        public override function begin(game:GameInterface):void
        {
            super.begin(game);

            piecesCollected = [false, false, false, false];

            layer = game.layers.addLayer();

            var w:Number = game.stageWidth;
            var h:Number = game.stageHeight;

            // tile map
            tileMap = layer.add(TileMapWithUserData) as TileMapWithUserData;
            tileMap.x = 50;
            tileMap.y = 50;
            tileMap.checkMapEdgeCollisions = false;

            // player
            player = layer.add(Player) as Player;
            player.x = 0.50 * w;
            player.y = 0.50 * h;

            // border
            var border:GameObject = layer.add();
            border.graphics.beginFill(0);
            border.graphics.drawRect(0,0,w,50); // top
            border.graphics.drawRect(0,50,50,h-100); // left
            border.graphics.drawRect(w-50,50,50,h-100); // right
            border.graphics.drawRect(0,h-50,w,50); // bottom

            // noise
            noise = layer.add(CanvasGameObject) as CanvasGameObject;
            noise.x = 50;
            noise.y = 50;
            noise.width = 0.3334 * (game.stageWidth - 100);
            noise.height = 0.3334 * (game.stageHeight - 100);
            noise.scaleX = 3;
            noise.scaleY = 3;
            noise.alpha = NOISE_ALPHA;

            // room
            roomRow = RoomData.START_ROOM_ROW;
            roomCol = RoomData.START_ROOM_COL;
            roomLoader = new RoomLoader();
            roomLoader.loadRoom(roomRow, roomCol, tileMap);

            game.cheats.addCheatCode("towlr", towlrCheat);
        }

        override public function end():void
        {
            super.end();
            game.layers.removeAll();
            game.cheats.removeCheatCode("towlr", towlrCheat);
        }

        private function towlrCheat():void
        {
            player.color = Player.RAINBOW;
            towlrReset();
            roomRow = 8;
            roomCol = 3;
            roomLoader.loadRoom(roomRow, roomCol, tileMap);
            postLoadCheck();
        }

        private function redoNoise():void
        {
            noise.canvas.bitmapData.noise(Math.random() * int.MAX_VALUE, 0, 256, 7, true);
        }


        /**
         * called every frame to check collisions
         */
        override public function collision():void
        {
            super.collision();
            tileMap.collide(player, game.time.timeDelta, tileCollisionCallback);
        }

        /**
         * per-frame updating
         */
        override public function update():void
        {
            super.update();

            checkForRoomChange();
            redoNoise();
        }

        /**
         * see if he got outside the level
         */
        protected function checkForRoomChange():void
        {
            if (player.y < tileMap.y) { roomChangeUp(); }
            else if (player.x > tileMap.x + tileMap.width) { roomChangeRight(); }
            else if (player.y > tileMap.y + tileMap.height) { roomChangeDown(); }
            else if (player.x < tileMap.x) { roomChangeLeft(); }
        }

        /**
         * move over one room
         */
        protected function roomChangeUp():void
        {
            roomRow = MathUtil.wrap(roomRow-1, 0, RoomData.ROOMS.length);
            player.y = tileMap.y + tileMap.height - 1;
            roomLoader.loadRoom(roomRow, roomCol, tileMap);
            postLoadCheck();
        }

        /**
         * move over one room
         */
        protected function roomChangeRight():void
        {
            roomCol = MathUtil.wrap(roomCol+1, 0, RoomData.ROOMS[roomRow].length);
            player.x = tileMap.x + 1;
            roomLoader.loadRoom(roomRow, roomCol, tileMap);
            postLoadCheck();
        }

        /**
         * move over one room
         */
        protected function roomChangeDown():void
        {
            roomRow = MathUtil.wrap(roomRow+1, 0, RoomData.ROOMS.length);
            player.y = tileMap.y + 1;
            roomLoader.loadRoom(roomRow, roomCol, tileMap);
            postLoadCheck();
        }

        /**
         * move over one room
         */
        protected function roomChangeLeft():void
        {
            roomCol = MathUtil.wrap(roomCol-1, 0, RoomData.ROOMS[roomRow].length);
            player.x = tileMap.x + tileMap.width - 1;
            roomLoader.loadRoom(roomRow, roomCol, tileMap);
            postLoadCheck();
        }

        /**
         * loop over every tile in the room and see if we need to mess with it
         */
        protected function postLoadCheck():void
        {
            var data:Array = tileMap.getAllTileData();
            for (var row:uint = 0; row < data.length; ++row)
            {
                var rowData:Array = data[row];
                if (!rowData) { continue; }
                for (var column:uint = 0; column < rowData.length; ++column)
                {
                    var tile:Object = rowData[column];
                    if (null == tile) { continue; }

                    // collected pieces
                    var fixPiece:Boolean = false;
                    switch (tile.userdata)
                    {
                        case "piece0":
                            if (piecesCollected[0])
                            {
                                fixPiece = true;
                            }
                            break
                        case "piece1":
                            if (piecesCollected[1])
                            {
                                fixPiece = true;
                            }
                            break
                        case "piece2":
                            if (piecesCollected[2])
                            {
                                fixPiece = true;
                            }
                            break
                        case "piece3":
                            if (piecesCollected[3])
                            {
                                fixPiece = true;
                            }
                            break
                    }
                    if (fixPiece)
                    {
                        tileMap.setTileWithUserdata(tile.col, tile.row, tile.index + 1, tile.userdata + "ghost");
                    }
                }
            }
        }

        /**
         * react to the player hitting a tile
         */
        protected function tileCollisionCallback(tileData:Object, tileMap:TileMapWithUserData, player:Player, ci:CollisionInfo):void
        {
            var index:int = tileData.index;
            if (index < 0) {return;}
            var ts:TileSet = roomLoader.getTileSet(tileMap);
            var userdata:String = ts.userdata[index];

            // wall breaking
            if (null != userdata)
            {
                var canBreakWall:Boolean = ((player.color == Player.RAINBOW) && StringUtil.endsWith(userdata, "wall") && (userdata!="towlr-wall")) ||
                                           ((player.color == Player.RED)   && (userdata == "redwall") ) ||
                                           ((player.color == Player.GREEN) && (userdata == "greenwall") ) ||
                                           ((player.color == Player.BLUE)  && (userdata == "bluewall") );
                if (canBreakWall)
                {
                    Sfx.destroyWall();
                    breakWall(tileData.row, tileData.col, userdata);
                    if (player.color != Player.RAINBOW)
                    {
                        player.color = Player.WHITE;
                    }
                    // unbounce
                    if (Math.abs(ci.normal.x) > Math.abs(ci.normal.y))
                    {
                        player.dx = - player.dx;
                    }
                    else
                    {
                        player.dy = -player.dy;
                    }
                }
            }

            // collecting pieces
            switch (userdata)
            {
                case "piece0":
                    collectedPiece(0, tileData, tileMap);
                    Sfx.collectPiece();
                    break;
                case "piece1":
                    collectedPiece(1, tileData, tileMap);
                    Sfx.collectPiece();
                    break;
                case "piece2":
                    collectedPiece(2, tileData, tileMap);
                    Sfx.collectPiece();
                    break;
                case "piece3":
                    collectedPiece(3, tileData, tileMap);
                    Sfx.collectPiece();
                    break;
            }

            // TOWLR
            if (userdata == "towlr-wall")
            {
                towlrReset();
                noise.animate( { alpha:1.0 }, { alpha:NOISE_ALPHA }, 0.25);
                Sfx.towlrFail();
            }

            // CAKE
            if (userdata == "cake")
            {
                game.changeGameState(new Victory());
            }
        }

        protected function towlrReset():void
        {
            game.time.timeScale = 1;
            player.dx = 0;
            player.dy = 0;
            player.x = 0.5 * game.stageWidth;
            player.y = tileMap.y + 1;
        }

        protected function collectedPiece(index:uint, tileData:Object, tileMap:TileMapWithUserData):void
        {
            if (!piecesCollected[index])
            {
                piecesCollected[index] = true;
                tileMap.setTileWithUserdata(tileData.col, tileData.row, tileData.index + 1, tileData.userdata+"ghost")
            }

            var numCollected:int = 0;
            for each (var b:Boolean in piecesCollected)
            {
                if (b)
                {
                    numCollected += 1;
                }
            }
            if (numCollected >= 4)
            {
                player.color = Player.RAINBOW;
                scheduler.callOnce(4.0, explainRainbow);
            }

            addDisappearingTextField(collectPieceMessages[numCollected-1]);
        }

        protected function explainRainbow():void
        {
            addDisappearingTextField("RAINBOW MODE!!!");
        }

        /**
         * break the wall at the given location if its userdata matches the argument
         */
        protected function breakWall(row:uint, col:uint, targetUserdata:String):void
        {
            var userdata:String = tileMap.getUserdataAt(row, col);
            if (userdata != targetUserdata) { return; }

            tileMap.clearTile(col, row);

            breakWall(row, col-1, targetUserdata);
            breakWall(row, col+1, targetUserdata);
            breakWall(row-1, col, targetUserdata);
            breakWall(row+1, col, targetUserdata);
        }

        /**
         * add a disappearing text field to the scene
         */
        private function addDisappearingTextField(text:String, x_pct:Number = 0.5, y_pct:Number = 0.5, time:Number = 3.0, fadeTime:Number = 2.0, fontSize:Number=36, color:Number = Colors.WHITE_ENOUGH):DisappearingText
        {
            var disappearingText:DisappearingText = layer.add(DisappearingText) as DisappearingText;
            disappearingText.timeRemaining = time;
            disappearingText.fadeTime = fadeTime;
            disappearingText.setFontSize(fontSize);
            disappearingText.fontColor = color;
            disappearingText.setText(text);
            disappearingText.setFont(Title.TuffyFont);
            disappearingText.centerOn( x_pct * game.stageWidth, y_pct * game.stageHeight );

            return disappearingText;
        }
    }
}
