﻿package stego
{
    import flash.display.MovieClip;
    import flash.geom.Rectangle;
    import stego.script.step.AnimateTo;
    import stego.util.ArrayUtil;
    import stego.util.collision.CollisionInfo;
    import stego.util.collision.HitBox;
    import stego.util.ObjectPool;
    import stego.util.PhysicsUtil;
    import stego.util.scene.GameObjectHolder;
    import stego.util.schedule.Schedule;
    import stego.behavior.IBehavior;

    /**
     * ...
     * @author Mike Meyer
     */
    public class GameObject extends MovieClip
    {
        /**
         * rate of movement in x.  Note that if you change this directly, you
         * get discontinuous acceleration.  That's great for jumps and
         * collisions, but if you want something smooth like gravity, set ddx
         * instead.
         */
        public var dx:Number = 0.0;

        /**
         * rate of movement in y.  Note that if you change this directly, you
         * get discontinuous acceleration.  That's great for jumps and
         * collisions, but if you want something smooth like gravity, set ddy
         * instead.
         */
        public var dy:Number = 0.0;

        /**
         * rate of acceleration in x.
         * Note that this gets cleared every frame, so you have to
         * continously add to it if you want something like gravity or wind
         * Also note that this is smoothed, which is
         * great for gravity but not for discontinuous movement like
         * collisions or jumps.  use dx for those
         */
        public var ddx:Number = 0.0;

        /**
         * rate of acceleration in y.
         * Note that this gets cleared every frame, so you have to
         * continously add to it if you want something like gravity or wind
         * Also note that this is smoothed, which is
         * great for gravity but not for discontinuous movement like
         * collisions or jumps.  use dy for those
         */
        public var ddy:Number = 0.0;

        public var dr:Number = 0.0;

        /// workaround for Flash rounding to twips
        private var _x:Number = 0.0;
        public override function get x () : Number { return _x; }
        public override function set x (value:Number) : void { _x = value; super.x = value; }

        /// workaround for Flash rounding to twips
        private var _y:Number = 0.0;
        public override function get y () : Number { return _y; }
        public override function set y (value:Number) : void { _y = value; super.y = value; }

        /// so that physics utilities can tell what they need to ignore
        public var physicsEnabled:Boolean;

        /// whether to add global gravity to the ddy every frame
        public var affectedByGravity:Boolean = false;

        public var alive:Boolean;

        private var _behaviors:Array;

        /**
         * whether the object is resting on the ground
         * invalid while inside PhysicsUtil.update, so don't try to use it then.
         */
        public var isOnGround:Boolean;

        // check to see if a game object is on screen
        public function get isOnScreen():Boolean
        {
            var rect:Rectangle = getBounds(parent);
            return    (rect.left   < stage.stageWidth)
                   && (rect.right  > 0)
                   && (rect.top    < stage.stageHeight)
                   && (rect.bottom > 0) ;
        }

        /** scheduling utility used for animations */
        protected var _scheduler:Schedule;

        /**
         * GameObjectHolder for managing internal gameobjects
         */
        private var _holder:GameObjectHolder;
        public function get internalGameObjects():GameObjectHolder { return _holder; }

        /** the game it belongs to */
        public var game:GameInterface;

        /** maximum fall speed (assuming +y is down) enforced in integrate */
        public var maxFallSpeed:Number = Number.POSITIVE_INFINITY;

        /**
         * initial setup and preparation for reuse
         */
        public function init(game:GameInterface):void
        {
            x = 0;
            y = 0;
            dx = 0;
            dy = 0;
            ddx = 0;
            ddy = 0;
            this.game = game;
            visible = true;
            physicsEnabled = true;
            affectedByGravity = false;
            if (null == _holder)
            {
                _holder = new GameObjectHolder(this, game);
            }
            alive = true;
        }

        public function integrate(elapsedTime:Number):void
        {
            enforceSpeedLimits();

            x += dx * 0.5 * elapsedTime;
            y += dy * 0.5 * elapsedTime;
            dx += ddx * elapsedTime;
            dy += ddy * elapsedTime;

            enforceSpeedLimits();

            x += dx * 0.5 * elapsedTime;
            y += dy * 0.5 * elapsedTime;
            ddx = 0;
            ddy = 0;

            rotation += dr * elapsedTime;

            internalGameObjects.integrate(elapsedTime);
        }

        protected function enforceSpeedLimits():void
        {
            dy = Math.min(dy, maxFallSpeed);
        }


        /**
         * per-frame updates
         * @param	elapsedTime - how long the last frame was
         */
        public function update(elapsedTime:Number):void
        {
            internalGameObjects.update(elapsedTime);

            if (null != _behaviors)
            {
                for each (var b:IBehavior in _behaviors)
                {
                    b.update(elapsedTime);
                }
            }

            if (_scheduler)
            {
                _scheduler.update(elapsedTime);
            }
        }

        /**
         * called after all the final positions have been figured
         * just before drawing
         */
        public function preDraw(timeDelta:Number):void
        {
            internalGameObjects.preDraw(timeDelta);

            // right before we go back into update seems like the right time to do this
            if (game && game.physics && affectedByGravity)
            {
                ddy += game.physics.gravity;
            }
        }

        /**
         * get this gameobject's hitboxes.
         * The Array returned by this can be returned with stego.util.ObjectPool.returnArray(a, stego.util.collision.HitBox)
         * Uses getBounds by default! Override if you do your own hitboxes!
         */
        public function getHitboxes():Array
        {
            var a:Array = ObjectPool.getArray(HitBox, 1);
            var hb:HitBox = a[0];
            hb.userData = "";
            hb.color = 0xFFFFFF;

            var bounds:Rectangle = getBounds(parent);
            hb.x = bounds.x - x;
            hb.y = bounds.y - y;
            hb.width = bounds.width;
            hb.height = bounds.height;

            return a;
        }

        /**
         * cleanup for when we are done with this instance for a while.
         * May still get reused!
         */
        public function shutdown():void
        {
            if (_scheduler)
            {
                _scheduler.shutdown();
                ObjectPool.returnObject(Schedule, _scheduler);
                _scheduler = null;
            }
            removeAllBehaviors();
        }

        /**
         * override to deal with collisions
         */
        public function collisionOccurred(ci:CollisionInfo, object:GameObject):void
        {
            if (!canPassThrough(ci, object))
            {
                var collisionBox:Rectangle = ci.hitbox1;

                var collision_dx:Number = (null != object) ? object.dx : 0;
                var collision_dy:Number = (null != object) ? object.dy : 0;

                if ( ci.normal.x > 0.0 )
                {
                    x = ci.location.x - collisionBox.left;
                    if (dx < collision_dx) { dx = collision_dx; }
                    if (ddx < 0) { ddx = 0; }
                }
                else if ( ci.normal.x < 0.0 )
                {
                    x = ci.location.x - collisionBox.right;
                    if (dx > collision_dx) { dx = collision_dx; }
                    if (ddx > 0) { ddx = 0; }
                }

                if ( ci.normal.y > 0.0  )
                {
                    y = ci.location.y - collisionBox.top;
                    if (dy < collision_dy) { dy = collision_dy; }
                    if (ddy < 0) { ddy = 0; }
                }
                else if ( ci.normal.y < 0.0 )
                {
                    y = ci.location.y - collisionBox.bottom;
                    if (dy > collision_dy) { dy = collision_dy; }
                    if (ddy > 0) { ddy = 0; }
                }

                isOnGround = isOnGround || (ci.normal.y < -Math.SQRT1_2);
            }
        }

        /**
         * override to let yourself go through things
         */
        public function canPassThrough(ci:CollisionInfo, object:GameObject):Boolean
        {
            return false;
        }

        /**
         * start an animation
         * @param	startValues - where it should start at.  example: {x:100}
         * @param	endValues - where it is animating to.  example: {x:200}
         * @param	animDuration - how long it takes to run the animation
         */
        public function animate(startValues:Object, endValues:Object, animDuration:Number=1.0):void
        {
            if (!_scheduler) { _scheduler = ObjectPool.getObject(Schedule); }
            _scheduler.runScript(new stego.script.step.Animate(this, startValues, endValues, animDuration));
        }

        /**
         * start an animation
         * @param	objectToAnimate - what we're animating
         * @param	endValues - where it is animating to.  example: {x:200}
         * @param	animDuration - how long it takes to run the animation
         */
        public function animateTo(endValues:Object, animDuration:Number=1.0):void
        {
            if (!_scheduler) { _scheduler = ObjectPool.getObject(Schedule); }
            _scheduler.runScript(new stego.script.step.AnimateTo(this, endValues, animDuration));
        }

        public function addBehavior(value:IBehavior):void
        {
            if (null == _behaviors)
            {
                _behaviors = new Array();
            }
            _behaviors.push(value);
            value.character = this;
        }
        public function removeBehavior(value:IBehavior):void
        {
            ArrayUtil.removeEqual(_behaviors, value);
            if (value)
            {
                value.character = null;
            }
        }
        public function removeAllBehaviors():void
        {
            if (_behaviors) { _behaviors.length = 0; }
        }

        /**
         * Setup that needs to happen before we do the collision checks
         */
        public function preCollision(timeDelta:Number):void
        {
            isOnGround = false;

            internalGameObjects.preCollision(timeDelta);
        }

        /**
         * do collision checks on internal stuff
         */
        public function collision(timeDelta:Number):void
        {
            internalGameObjects.preCollision(timeDelta);
        }
    }
}
