﻿package stego.util
{
    import stego.collections.LinkedList;
    import stego.collections.LinkedListNode;
    import stego.GameInterface;
    import stego.GameObject;
    import stego.util.collision.CollisionInfo;
    import stego.util.collision.CollisionUtil;
    import stego.util.collision.HitBox;
    import stego.util.scene.GameObjectHolder;
    import stego.util.scene.GameObjectHolder;

    /**
     * Class to handle the physics simulation and collision checks.
     * to use it, add layers and collisionPairs to tell it what layers
     * get processed and what ones collide with each other.
     * @author Mike Meyer
     */
    public class PhysicsUtil
    {
        /**
         * list of pairs whose objects will get physics'd
         */
        private var _collisionPairs:/*Pair*/LinkedList;

        /**
         * get/set the acceleration due to gravity in this game
         */
        public var gravity:Number = 2000;

        /**
         * get/set how far things will snap left or right for collisions
         */
        public var horizontalSnap:Number = 0;

        /**
         * get/set how far things will snap up or down for collisions
         */
        public var verticalSnap:Number = 0;

        /**
         * game instance that this belongs to
         */
        private var _game:GameInterface;

        /**
         * set up the physics util
         */
        public function init(game:GameInterface):void
        {
            _game = game;
            _collisionPairs = new LinkedList();
        }
        /**
         * clean up the physics util
         */
        public function shutdown():void
        {
            _collisionPairs.clear();
            _game = null;
        }

        /**
         *
         */
        public function update(timeDelta:Number):void
        {
        }

        /**
         *
         */
        public function integrate(timeDelta:Number):void
        {
        }

        /**
         *
         */
        public function preCollision(timeDelta:Number):void
        {
        }

        /**
         *
         */
        public function collision(timeDelta:Number):void
        {
            var layer1Node:LinkedListNode;
            var layer1:GameObjectHolder;
            var object1Node:LinkedListNode;
            var object1:GameObject;
            var layer2:GameObjectHolder;
            var object2Node:LinkedListNode;
            var object2:GameObject;

            // check pairs for collisions
            for (var pairNode:LinkedListNode = _collisionPairs.first; pairNode != null; pairNode = pairNode.next)
            {
                var pair:Pair = pairNode.value;
                layer1 = pair.a;
                layer2 = pair.b;
                for (object1Node = layer1.objects.first; object1Node!=null; object1Node = object1Node.next)
                {
                    object1 = object1Node.value;
                    if (!object1.physicsEnabled) { continue; }

                    // don't double-compare if it is the same layer
                    object2Node = (layer1==layer2) ? object1Node.next : layer2.objects.first;

                    for (; object2Node != null; object2Node = object2Node.next)
                    {
                        object2 = object2Node.value;
                        if (!object2.physicsEnabled) { continue; }
                        checkCollision(object1, object2);
                    }
                }
            }
        }

        /**
         *
         */
        public function preDraw(timeDelta:Number):void
        {
        }

        /**
         * Detect and react to the collisions.
         * (stops at the first hitbox overlap it detects)
         * If allowedUserData is provided, only hitboxes whose userdata string
         * appears in that array will be compared.
         */
        protected function checkCollision(go1:GameObject, go2:GameObject, allowedUserData:Array = null):void
        {
            var ci:CollisionInfo = CollisionUtil.checkAabbCollision(go1, go2, _game.time.timeDelta, horizontalSnap, verticalSnap, allowedUserData);

            if (null != ci)
            {
                // let the objects know they hit each other
                go1.collisionOccurred(ci, go2);
                ci.normal.x *= -1;
                ci.normal.y *= -1;
                var temp:HitBox = ci.hitbox2;
                ci.hitbox1 = ci.hitbox2;
                ci.hitbox2 = temp;
                go2.collisionOccurred(ci, go1);
            }
        }

        /**
         * tell the physicsUtil to check collisions between objects in a and
         * objects in b.  (they can be the same layer)
         * Note that the collision code prefers to move A
         */
        public function addCollisionPair(a:GameObjectHolder, b:GameObjectHolder):void
        {
            _collisionPairs.push_back(new Pair(a, b));
        }

        /**
         * remove a collision pair
         */
        public function removeCollisionPair(a:GameObjectHolder, b:GameObjectHolder):void
        {
            var node:LinkedListNode = _collisionPairs.first,
                next:LinkedListNode,
                p:Pair;
            while (node)
            {
                next = node.next;
                p = node.value;

                if ( (p.a == a) && (p.b == b) )
                {
                    _collisionPairs.remove(node);
                }

                node = next;
            }
        }

        /**
         * remove all registered collision pairs
         */
        public function removeAllCollisionPairs():void
        {
            _collisionPairs.clear();
        }
    }

}
import stego.util.scene.GameObjectHolder;
import stego.util.scene.GameObjectHolder;

/**
 * for associating pairs of layers to do collision checks on
 */
class Pair
{
    public function Pair(a:GameObjectHolder = null, b:GameObjectHolder = null) { this.a = a; this.b = b; }
    public var a:GameObjectHolder;
    public var b:GameObjectHolder;
}
