﻿package stego.util.collision
{
    import flash.geom.Point;
    import flash.geom.Rectangle;
    import stego.GameObject;
    import stego.util.collision.CollisionInfo;
    import stego.util.MathUtil;
    import stego.util.ObjectPool;

    /**
     * handy functions for doing collision checks
     * @author Mike Meyer
     */
    public class CollisionUtil
    {
        // the farthest (relative) distance we want to go between collision tests
        private static const COLLISION_STEP_SIZE_MAX:Number = 1;

        /** avoid new'ing */
        protected static var _tempHitBox1:HitBox = new HitBox();
        protected static var _tempHitBox2:HitBox = new HitBox();
        protected static var _tempPoint1:Point = new Point();
        protected static var _tempPoint2:Point = new Point();
        protected static var _tempPoint3:Point = new Point();

        /**
         * check for collisions between two moving GameObjects
         * returns null if they did not collide
         * @param	objectA - first object in the collision-check pair
         * @param	objectB - second object in the collision-check pair
         * @param	elapsedTime - length of the time step in which we are checking collisions
         * @param	horizontalSnap - snap object A horizontally up to this difference if it would make the difference between a collision of not
         * @param	verticalSnap - snap object A vertically up to this difference if it would make the difference between a collision of not
         * @return  A CollisionInfo if there was a collision.
         */
        public static function checkAabbCollision( objectA:GameObject,
                                                   objectB:GameObject,
                                                   elapsedTime:Number,
                                                   horizontalSnap:Number,
                                                   verticalSnap:Number,
                                                   allowedUserData:Array = null ):CollisionInfo
        {
            var ci:CollisionInfo;
            var aHitboxes:Array = objectA.getHitboxes();
            var bHitboxes:Array = objectB.getHitboxes();
            for each (var objectAHitBox:HitBox in aHitboxes)
            {
                if (allowedUserData && -1==allowedUserData.indexOf(objectAHitBox.userData)) continue;
                for each (var objectBHitBox:HitBox in bHitboxes)
                {
                    if (allowedUserData && -1==allowedUserData.indexOf(objectBHitBox.userData)) continue;

                    ci = checkHitboxCollision( objectAHitBox,
                                               objectA,
                                               objectBHitBox,
                                               objectB,
                                               elapsedTime,
                                               horizontalSnap,
                                               verticalSnap,
                                               allowedUserData );
                }
            }

            ObjectPool.returnArray(aHitboxes, HitBox);
            ObjectPool.returnArray(bHitboxes, HitBox);

            return ci;
        }

        /**
         * check for collisions between the specified hitboxes of the specified objects
         * returns null if they did not collide
         * @param   hitboxA - hitbox of object A
         * @param	objectA - first object in the collision-check pair
         * @param   hitboxB - hitbox of object B
         * @param	objectB - second object in the collision-check pair
         * @param	elapsedTime - length of the time step in which we are checking collisions
         * @param	horizontalSnap - snap object A horizontally up to this difference if it would make the difference between a collision of not
         * @param	verticalSnap - snap object A vertically up to this difference if it would make the difference between a collision of not
         * @return  A CollisionInfo if there was a collision.
         */
        public static function checkHitboxCollision( hitboxA:HitBox,
                                                     objectA:GameObject,
                                                     hitboxB:HitBox,
                                                     objectB:GameObject,
                                                     elapsedTime:Number,
                                                     horizontalSnap:Number,
                                                     verticalSnap:Number,
                                                     allowedUserData:Array = null ):CollisionInfo
        {
            if (allowedUserData)
            {
                if ( ( -1==allowedUserData.indexOf(hitboxA.userData)) || ( -1==allowedUserData.indexOf(hitboxB.userData))  )
                {
                    return null;
                }
            }

            var avg_a_dx:Number = objectA.dx + 0.5 * objectA.ddx;
            var avg_b_dx:Number = objectB.dx + 0.5 * objectB.ddx;
            var avg_a_dy:Number = objectA.dy + 0.5 * objectA.ddy;
            var avg_b_dy:Number = objectB.dy + 0.5 * objectB.ddy;
            var aMotionRelativeToB_x:Number = (avg_a_dx - avg_b_dx) * elapsedTime;
            var aMotionRelativeToB_y:Number = (avg_a_dy - avg_b_dy) * elapsedTime;
            var ci:CollisionInfo = null;

            _tempPoint1.x = objectA.x;
            _tempPoint1.y = objectA.y;
            _tempPoint2.x = objectB.x;
            _tempPoint2.y = objectB.y;

            if (quickCollisionPossibilityCheck( hitboxA,
                                                _tempPoint1,
                                                hitboxB,
                                                _tempPoint2,
                                                aMotionRelativeToB_x,
                                                aMotionRelativeToB_y,
                                                horizontalSnap,
                                                verticalSnap) )
            {
                ci = CollisionUtil.internalCheckAabbCollision(
                    hitboxA,
                    _tempPoint1,
                    hitboxB,
                    _tempPoint2,
                    aMotionRelativeToB_x,
                    aMotionRelativeToB_y,
                    horizontalSnap,
                    verticalSnap);
            }
            return ci;
        }

        /**
         * check for collisions between two rects
         * @param   a           hitbox for one of the objects
         * @param   a_offset    translation to apply to a before checking for overlap
         * @param   b           hitbox for the other
         * @param   b_offset  translation to apply to a before checking for overlap
         * @param   leftSnap    how far left it can shift a to make a collision happen
         * @param   rightSnap   how far right it can shift a to make a collision happen
         * @param   upSnap      how far up it can shift a to make a collision happen
         * @param   downSnap    how far down it can shift a to make a collision happen
         */
        public static function checkRectCollision( a:HitBox,
                                                   a_offset:Point,
                                                   b:HitBox,
                                                   b_offset:Point,
                                                   leftSnap:Number = 0,
                                                   rightSnap:Number = 0,
                                                   upSnap:Number = 0,
                                                   downSnap:Number = 0 ):CollisionInfo
        {
            var a2:HitBox = _tempHitBox1;
            a2.makeIntoCopyOf(a);
            a2.offsetPoint(a_offset);
            var b2:HitBox = _tempHitBox2;
            b2.makeIntoCopyOf(b);
            b2.offsetPoint(b_offset);

            var hitInX:Boolean = (b2.left < a2.right) && (a2.left < b2.right),
                hitInY:Boolean = (b2.top < a2.bottom) && (a2.top < b2.bottom),
                snapInX:Boolean = (!hitInX) && ((b2.left - rightSnap) < a2.right) && (a2.left < (b2.right + leftSnap)),
                snapInY:Boolean = (!hitInY) && ((b2.top - downSnap) < a2.bottom) && (a2.top < (b2.bottom + upSnap));

            // out now if it's no good (at least one axis has to overlap.  we don't snap both.
            if (!hitInX && !hitInY)
            {
                return null;
            }

            var aCenterX:Number = a2.left + (0.5 * a2.width),
                aCenterY:Number = a2.top + (0.5 * a2.height),
                bCenterX:Number = b2.left + (0.5 * b2.width),
                bCenterY:Number = b2.top + (0.5 * b2.height),
                xOverlap:Number = 0,
                yOverlap:Number = 0;

            if (hitInX)
            {
                if ( (a2.left <= b2.left) && (b2.right <= a2.right) )
                {
                    xOverlap = b2.width;
                }
                else if ( (a2.left >= b2.left) && (b2.right <= a2.right) )
                {
                    xOverlap = Math.min(b2.right - a2.left, a2.right - b2.left);
                }
                else if (a2.left <= b2.left)
                {
                    xOverlap = a2.right - b2.left;
                }
                else
                {
                    xOverlap = a2.right - a2.left;
                }
            }

            if (hitInY)
            {
                if ( (a2.top <= b2.top) && (b2.bottom <= a2.bottom) )
                {
                    yOverlap = b2.width;
                }
                else if ( (a2.top >= b2.top) && (b2.bottom <= a2.bottom) )
                {
                    yOverlap = Math.min(b2.bottom - a2.top, a2.bottom - b2.top);
                }
                else if (a2.top <= b2.top)
                {
                    yOverlap = a2.bottom - b2.top;
                }
                else
                {
                    yOverlap = a2.bottom - a2.top;
                }
            }
            var xCollision:Boolean = hitInY && (xOverlap < a2.width) &&
                                     ( (hitInX && (xOverlap < yOverlap)) ||
                                       snapInX );
            var yCollision:Boolean = !xCollision && hitInX && (hitInY || snapInY);

            var ci:CollisionInfo = null;
            if (yCollision)
            {
                ci = new CollisionInfo();
                ci.location.x = aCenterX;
                ci.location.y = (aCenterY > bCenterY) ? b2.bottom : b2.top;
                ci.normal.x = 0;
                ci.normal.y = (aCenterY > bCenterY) ? 1.0 : -1.0;
                ci.hitbox1 = a;
                ci.hitbox2 = b;
            }
            else if (xCollision)
            {
                ci = new CollisionInfo();
                ci.location.x = (aCenterX > bCenterX) ? b2.right : b2.left;
                ci.location.y = aCenterY;
                ci.normal.x = (aCenterX > bCenterX) ? 1.0 : -1.0;
                ci.normal.y = 0;
                ci.hitbox1 = a;
                ci.hitbox2 = b;
            }

            return ci;
        }

        /**
         * check for collisions between two potentially moving rects
         */
        private static function internalCheckAabbCollision(
            objectAHitBox:HitBox,
            objectAOffset:Point,
            objectBHitBox:HitBox,
            objectBOffset:Point,
            aMotionRelativeToB_x:Number,
            aMotionRelativeToB_y:Number,
            horizontalSnap:Number,
            verticalSnap:Number ):CollisionInfo
        {
            var ci:CollisionInfo = null;
            var dist:Number = Math.sqrt( (aMotionRelativeToB_x*aMotionRelativeToB_x) + (aMotionRelativeToB_y*aMotionRelativeToB_y) );
            var steps:Number = 1 + Math.floor(dist / COLLISION_STEP_SIZE_MAX);

            for (var step:Number = 0; step <= steps; ++step)
            {
                var scale:Number = (step) / steps;
                _tempPoint3.x = objectAOffset.x + aMotionRelativeToB_x * scale;
                _tempPoint3.y = objectAOffset.y + aMotionRelativeToB_y * scale;

                ci = checkRectCollision( objectAHitBox,
                                         _tempPoint3,
                                         objectBHitBox,
                                         objectBOffset,
                                         (aMotionRelativeToB_x <= 0)?horizontalSnap:0,
                                         (aMotionRelativeToB_x >= 0)?horizontalSnap:0,
                                         (aMotionRelativeToB_y <= 0)?verticalSnap:0,
                                         (aMotionRelativeToB_y >= 0)?verticalSnap:0 );
                if (ci)
                {
                    ci.t = step / steps;
                    break;
                }
            }
            return ci;
        }

        /**
         * quick check for whether there is any chance of collision
         */
        private static function quickCollisionPossibilityCheck(
            objectABounds:Rectangle,
            objectAOffset:Point,
            objectBBounds:Rectangle,
            objectBOffset:Point,
            aMotionRelativeToB_x:Number,
            aMotionRelativeToB_y:Number,
            horizontalSnap:Number,
            verticalSnap:Number ):Boolean
        {
            // too far left
            var aMaxX:Number = Math.max(objectABounds.right, objectABounds.right + aMotionRelativeToB_x);
            if (objectAOffset.x + aMaxX < objectBOffset.x + objectBBounds.left - horizontalSnap) { return false; }

            // too far right
            var aMinX:Number = Math.min(objectABounds.left, objectABounds.left + aMotionRelativeToB_x);
            if (objectAOffset.x + aMinX > objectBOffset.x + objectBBounds.right + horizontalSnap) { return false; }

            // too far up
            var aMaxY:Number = Math.max(objectABounds.bottom, objectABounds.bottom + aMotionRelativeToB_y);
            if (objectAOffset.y + aMaxY < objectBOffset.y + objectBBounds.top - verticalSnap) { return false; }

            // too far down
            var aMinY:Number = Math.min(objectABounds.top, objectABounds.top + aMotionRelativeToB_y);
            if (objectAOffset.y + aMinY > objectBOffset.y + objectBBounds.bottom + verticalSnap) { return false; }

            return true;
        }
    }
}
