package stego.util.graphics
{
    import flash.display.Bitmap;
    import flash.display.BitmapData;
    import stego.util.MathUtil;

    /**
     * utility to help draw animations out of a spritesheet
     */
    public class AnimatedSpriteSheet extends SpriteSheet
    {
        /**
         * lets you control the playback speed of animations
         * 1.0 is normal speed, 2.0 is twice as fast, 0.1 is ten times as slow, etc.
         */
        public var speed:Number = 1.0;

        /** get the currently playing animation's name */
        public function get currentAnim():String { return _currentAnim; }
        protected var _currentAnim:String = "";

        /** start/stop playback */
        public var paused:Boolean = false;

        /** animation data indexed by the anim name */
        protected var _animations:Object;

        /** current time into the animation */
        protected var _animTimer:Number = 0;

        /** callback for when the animation finishes or is cancelled */
        protected var callback:Function;

        /**
         * constructor
         */
        public function AnimatedSpriteSheet(target:Bitmap = null, source:BitmapData = null, spriteWidth:int = 100, spriteHeight:int = 100):void
        {
            super(target, source, spriteWidth, spriteHeight);
            _animations = { };
        }

        /**
         * clear out any animation data this has
         */
        public function removeAllAnimations():void
        {
            _animations = { };
        }

        /**
         * add an animation
         * @param	animName    name to refer to it by
         * @param	frames      array listing out which frames appear in the animation in what order
         * @param	rate        how fast the animation plays. (frames per second)
         * @param	loop        whether the animation should loop or stop at the end
         */
        public function addAnimation(name:String, frames:Array, rate:Number, loop:Boolean):void
        {
            _animations[name] = new Animation(frames, rate, loop);
        }

        /**
         * animate!
         */
        public function update(timeDelta:Number):void
        {
            if (_animations.hasOwnProperty(_currentAnim))
            {
                var anim:Animation = _animations[_currentAnim];
                var numFrames:uint = anim.frames.length;
                var animLength:Number = numFrames / anim.rate;
                if (!paused)
                {
                    _animTimer += speed * timeDelta;
                }

                if (anim.loop)
                {
                    _animTimer = MathUtil.wrap(_animTimer, 0.0, animLength);
                }
                else
                {
                    _animTimer = MathUtil.clamp(_animTimer, 0, animLength);
                }

                // let the base class's setter do most of the work
                var frame:uint = _animTimer * anim.rate;
                if (frame > numFrames - 1)
                {
                    frame = numFrames - 1;

                    // finish callback gets called when we hit the end and are not looping
                    if ( (!anim.loop) && (null != callback) )
                    {
                        var temp:Function = callback;
                        callback = null;
                        temp(true);
                    }
                }
                currentFrame = anim.frames[frame];
            }
            else
            {
                _canvas.clear();
            }
        }

        /**
         * start playing an animation
         * @param   animName    animation to play.  should match a name you gave w/ addAnimation
         * @param   callback    function to call when the animation finishes or is cancelled
         *                      should look like
         *                      function foo(finished:Boolean):void
         */
        public function playAnimation(animName:String, callback:Function = null):void
        {
            if (null != this.callback) { callback(false); }
            this.callback = callback;

            _currentAnim = animName;

            _animTimer = 0;
            if (_animations.hasOwnProperty(animName))
            {
                var anim:Animation = _animations[animName];
                var rate:Number = anim.rate;
                if (rate * speed < 0)
                {
                    _animTimer = anim.frames.length / rate;
                }
            }
            else if (null != this.callback)
            {
                // don't forget to call the callback if we fail to start an animation
                var temp:Function = callback;
                this.callback = null;
                temp(false);
            }
        }

        /**
         * stop the animation and remove the callback.
         */
        public function cancel():void
        {
            if (null != this.callback)
            {
                var temp:Function = callback;
                this.callback = null;
                temp(false);
            }
            _currentAnim = null;
        }
    }
}

/**
 * helper class for storing information about an animation
 */
class Animation
{
    /** frame indices used in the animation in order */
    public var frames:Array;

    /** how fast it runs in frames per second */
    public var rate:Number;

    /** whether the animation should loop or stop at the end */
    public var loop:Boolean;

    /** constructor */
    public function Animation(frames:Array, rate:Number, loop:Boolean)
    {
        this.frames = frames;
        this.rate = rate;
        this.loop = loop;
    }
}
