﻿package stego.util.scene
{
    import flash.display.DisplayObjectContainer;
    import flash.utils.Dictionary;
    import stego.GameInterface;
    import stego.GameObject;
    import stego.collections.LinkedList;
    import stego.collections.LinkedListNode;

    /**
     * ...
     * @author Mike Meyer
     */
    public class GameObjectHolder
    {
        public var container:DisplayObjectContainer;

        protected var _liveObjects:/*GameObject*/LinkedList;
        public function get objects():/*GameObject*/LinkedList { return _liveObjects; }

        public var game:GameInterface;

        /** Lists of reusable game objects instanced by class. */
        public static var reusables:Dictionary = new Dictionary();

        /** whether game objects in this holder get update, integrate, etc. called on them */
        public var enabled:Boolean = true;

        /** number of times GameObjectHolder has had to new something */
        public static var numNews:uint = 0;

        /**
         * constructor
         * @param	container   what in the scene holds this
         * @param	game        game class this belongs to
         */
        public function GameObjectHolder(container:DisplayObjectContainer = null, game:GameInterface = null)
        {
            this.container = container;
            _liveObjects = new LinkedList(true);

            this.game = game;
        }

        /**
         * add a game object to the holder.  unlike Add, it expects you to
         * provide a new instance and does not reuse objects
         */
        public function addPremade(go:GameObject):void
        {
            addPremadeAtIndex(go, container.numChildren);
        }

        /**
         * add a game object to the holder.  unlike Add, it expects you to
         * provide a new instance and does not reuse objects
         */
        public function addPremadeAtIndex(go:GameObject, index:Number):void
        {
            if (null != go.parent)
            {
                throw new Error("Adding a GameObject that already has a parent!");
            }

            container.addChildAt(go, index);
            go.init(game);
            _liveObjects.push_back(go);
        }


        /// Add a game object to the holder.  Will try to re-use existing ones
        public function add(type:Class = null):GameObject
        {
            return addAtIndex(container.numChildren, type);
        }

        /// Add a game object to the holder.  Will try to re-use existing ones
        public function addAtIndex(index:Number, type:Class = null):GameObject
        {
            if (null == type)
            {
                type = stego.GameObject;
            }

            var go:GameObject = findReusableObject(type);
            if (null == go)
            {
                ++numNews;
                go = new type();
            }

            addPremadeAtIndex(go, index);
            return go;
        }

        private function findReusableObject(type:Class):GameObject
        {
            if (type in reusables)
            {
                if (0 < reusables[type].length)
                {
                    return reusables[type].pop();
                }
            }

            return null;
        }

        /**
         * "heartbeat" function. called every frame
         */
        public function update(elapsedTime:Number):void
        {
            if (!enabled) return;

            var node:LinkedListNode = _liveObjects.first;
            while (node)
            {
                var next:LinkedListNode = node.next;
                var go:GameObject = node.value;
                go.update(elapsedTime);
                node = next;
            }

            removeDeadObjects();
        }

        /**
         * called every frame to integrate object motion
         */
        public function integrate(timeDelta:Number):void
        {
            if (!enabled) return;

            var node:LinkedListNode = _liveObjects.first;
            while (node)
            {
                var next:LinkedListNode = node.next;
                var go:GameObject = node.value;
                go.integrate(timeDelta);
                node = next;
            }
        }

        /**
         * called every frame before collisions
         */
        public function preCollision(timeDelta:Number):void
        {
            if (!enabled) return;

            var node:LinkedListNode = _liveObjects.first;
            while (node)
            {
                var next:LinkedListNode = node.next;
                var go:GameObject = node.value;
                go.preCollision(timeDelta);
                node = next;
            }
        }

        /**
         * called every frame to check collisions
         */
        public function collision(timeDelta:Number):void
        {
            if (!enabled) return;

            var node:LinkedListNode = _liveObjects.first;
            while (node)
            {
                var next:LinkedListNode = node.next;
                var go:GameObject = node.value;
                go.collision(timeDelta);
                node = next;
            }
        }

        /**
         * last-minute updating before things get rendered each frame
         */
        public function preDraw(timeDelta:Number):void
        {
            if (!enabled) return;

            var node:LinkedListNode = _liveObjects.first;
            while (node)
            {
                var next:LinkedListNode = node.next;
                var go:GameObject = node.value;
                go.preDraw(timeDelta);
                node = next;
            }
        }

        private function removeDeadObjects():void
        {
            var node:LinkedListNode = _liveObjects.first;
            while (node)
            {
                var next:LinkedListNode = node.next;
                var go:GameObject = node.value;

                // if it dies. queue it up for removal
                if (!go.alive)
                {
                    remove(go);
                }
                node = next;
            }
        }

        /**
         * remove the game object from the scene.  Store it in
         * reusables for later reuse.
         */
        public function remove(go:GameObject):void
        {
            if (!go) { return; }
            go.shutdown();

            var goClass:Class = Object(go).constructor;

            // the base gameobject class will probably have custom stuff
            // attached in it. clear that stuff out.
            if (GameObject === goClass)
            {
                var i:int = go.numChildren;
                while( i-- )
                {
                    go.removeChildAt(i);
                }
                go.graphics.clear();
            }

            if (!(goClass in reusables))
            {
                reusables[goClass] = [];
            }
            reusables[goClass].push(go);

            if (go.parent == container)
            {
                container.removeChild(go);
            }
            _liveObjects.removeByValue(go);
        }

        /**
         * call a function for each go in the layer.  func should be of the form
         * function func(go:GameObject):void
         */
        public function forEach(func:Function):void
        {
            var node:LinkedListNode = _liveObjects.first
            while (node)
            {
                var next:LinkedListNode = node.next;
                var go:GameObject = node.value;
                func(go);
                node = next;
            }
        }

        /**
         * Get rid of any and all objects held by this thing
         */
        public function removeAll():void
        {
            var node:LinkedListNode = _liveObjects.first
            while (node)
            {
                var next:LinkedListNode = node.next;
                var go:GameObject = node.value;
                remove(go);
                node = next;
            }
        }

        /**
         * empty out our cache of reusable objects
         * This affects ALL GameObjectHolders
         */
        public static function clearReusables():void
        {
            GameObjectHolder.reusables = new Dictionary();
        }
    }

}
